/* 
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005/2006, Anthony Minessale II <anthmct@yahoo.com>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthmct@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Anthony Minessale II <anthmct@yahoo.com>
 * Neal Horman <neal at wanlink dot com>
 *
 *
 * mod_squirrelfish.c -- Framework Demo Module
 *
 */
#include <switch.h>
#include <JavaScriptCore/JavaScript.h>

SWITCH_MODULE_LOAD_FUNCTION(mod_squirrelfish_load);
SWITCH_MODULE_DEFINITION(mod_squirrelfish, mod_squirrelfish_load, NULL, NULL);

static JSValueRef print_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount,
		 const JSValueRef arguments[],
		 JSValueRef* exception);

static JSValueRef session_ready_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount,
		 const JSValueRef arguments[],
		 JSValueRef* exception);

/*
 * Functions
 */
typedef JSValueRef (*JSFunctionCB_t)(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
			size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception);

static struct globalFunction {
	const char	*name;
	JSFunctionCB_t	 func;
} globalFunctions[] = {
	{ "consoleLog",		&print_cb },
	{ "console_log",	&print_cb },
	{ "setGlobalVariable",	NULL },
	{ "getGlobalVariable",	NULL },
	{ "exit",		NULL },
	{ "include",		NULL },
	{ "bridge",		NULL },
	{ "apiExecute",		NULL },
	{ "use",		NULL },
	{ "fileDelete",		NULL },
	{ "system",		NULL },

#ifdef HAVE_CURL
	{ "fetchURL",		NULL },
	{ "fetchUrl",		NULL },
	{ "fetchURLHash",	NULL },
	{ "fetchUrlHash",	NULL },
	{ "fetchURLFile",	NULL },
	{ "fetchUrlFile",	NULL },
#endif

	{ 0, 0 }	/*!< don't delete! */
};

static switch_status_t init_functions(JSContextRef context)
{
	struct globalFunction *gfunc = globalFunctions;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name;
	JSObjectRef function;

	if(!context || !globalObject)
		return SWITCH_STATUS_GENERR;

	while(gfunc && gfunc->name) {
		if(gfunc->func) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Creating global function %s\n", gfunc->name);

			name = JSStringCreateWithUTF8CString(gfunc->name);
			function = JSObjectMakeFunctionWithCallback(context, name, gfunc->func);
			JSObjectSetProperty(context, globalObject, name, function, kJSPropertyAttributeNone, NULL);
			JSStringRelease(name);
		}
		gfunc++;
	}

	return SWITCH_STATUS_SUCCESS;
}


/******************************************************************************
 * Session class
 ******************************************************************************/

/*
 * Forward declarations
 */
//void sessionClass_Initialize(JSContextRef ctx, JSObjectRef object);
//void sessionClass_Finalize(JSObjectRef object);

bool sessionClass_HasProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName);
JSValueRef sessionClass_GetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception);
bool sessionClass_SetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef value, JSValueRef* exception);
//bool sessionClass_DeleteProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception);
//void sessionClass_GetPropertyNames(JSContextRef ctx, JSObjectRef object, JSPropertyNameAccumulatorRef propertyNames);

//JSValueRef sessionClass_CallAsFunction(JSContextRef ctx, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception);
//JSObjectRef sessionClass_CallAsConstructor(JSContextRef ctx, JSObjectRef constructor, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception);
//bool sessionClass_HasInstance(JSContextRef ctx, JSObjectRef constructor, JSValueRef possibleInstance, JSValueRef* exception);
//JSValueRef sessionClass_ConvertToType(JSContextRef ctx, JSObjectRef object, JSType type, JSValueRef* exception);


static JSStaticValue sessionClassStaticValues[] = {
	{ 0, 0, 0, 0 }
};

static JSStaticFunction sessionClassStaticFunctions[] = {
	{ "log",	&print_cb,		kJSPropertyAttributeNone },	/*!< core logger function */
//	{ "ready",	&session_ready_cb,	kJSPropertyAttributeNone },	/*!< session state */
	{ 0, 0, 0 }
};

static JSClassDefinition sessionClass_definition = {
	.version	= 0,
	.attributes	= kJSClassAttributeNone,

	.className	= "Session",
	.parentClass	= NULL,

	.staticValues	= sessionClassStaticValues,	/*!< static values */
	.staticFunctions= sessionClassStaticFunctions,	/*!< static members */

	.initialize		= NULL,		/*!< invoked during creation of the object */
	.finalize		= NULL,		/*!< called before garbage collecting this object */
	.hasProperty		= sessionClass_HasProperty,	/*!< check if object has a specific property */
	.getProperty		= sessionClass_GetProperty,	/*!< get value of property */
	.setProperty		= sessionClass_SetProperty,	/*!< set value of property */
	.deleteProperty		= NULL,		/*!< delete property */
	.getPropertyNames	= NULL,		/*!< return names of all properties */
	.callAsFunction		= NULL,		/*!< invoked if object has been called as function */
	.callAsConstructor	= NULL,		/*!< invoked if new has been called on class */
	.hasInstance		= NULL,		/*!< invoked on instanceof call */
	.convertToType		= NULL		/*!< convert object to string or integer */
};

static JSClassRef sessionClass;

/**
 * create_sessionObject
 */
static JSClassRef create_sessionClass(JSContextRef context)
{
	if(!sessionClass) {
		sessionClass = JSClassCreate(&sessionClass_definition);
	}
	return sessionClass;
}

/*
 * Session class callback functions
 */
static JSValueRef session_ready_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount,
		 const JSValueRef arguments[],
		 JSValueRef* exception)
{
	switch_core_session_t *session = JSObjectGetPrivate(thisObject);

	if(!session || !switch_channel_ready(switch_core_session_get_channel(session))) {
		return JSValueMakeBoolean(context, false);
	}

	return JSValueMakeBoolean(context, true);
}

/******************************************************************************
 *
 ******************************************************************************/

typedef JSClassRef (*createClassCB_t)(JSContextRef context);

static struct globalClass {
	const char	*name;
	createClassCB_t	 createfunc;
} globalClasses[] = {
	{ "Session", &create_sessionClass },
	{ 0, 0 }
};

/**
 * init_classes
 * \brief	Create global class objects (e.g. "Session")
 */
static switch_status_t init_classes(JSContextRef context)
{
	struct globalClass *gclass = globalClasses;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name;
	JSObjectRef object;

	if(!context || !globalObject)
		return SWITCH_STATUS_GENERR;

	while(gclass && gclass->name) {
		if(gclass->createfunc) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Creating global class %s\n", gclass->name);

			name = JSStringCreateWithUTF8CString(gclass->name);
			object = JSObjectMake(context, gclass->createfunc(context), NULL);
			JSObjectSetProperty(context, globalObject, name, object, kJSPropertyAttributeNone, NULL);
			JSStringRelease(name);
		}
		gclass++;
	}

	return SWITCH_STATUS_SUCCESS;
}

static struct objectMethod {
	const char	*name;
	JSFunctionCB_t	 func;
} sessionMethods[] = {
	{ "ready", &session_ready_cb },
	{ 0, 0 }
};


/*
 * Dynamic properties are handled via session_getProperty()
 * this array is used there
 */
typedef JSValueRef (*JSPropertyGetterCB_t)(JSContextRef, JSObjectRef);
typedef bool (*JSPropertySetterCB_t)(JSContextRef, JSObjectRef, JSValueRef value, JSValueRef *exception);

static struct objectProperty {
	const char		*name;
	JSPropertySetterCB_t	 setfunc;
	JSPropertyGetterCB_t	 getfunc;
	unsigned int	 	 flags;
} sessionProperties[] = {
	{ 0, 0, 0, 0 }
};

/**
 * sessionClass_HasProperty
 * \todo	Read-Only static properties?
 */
bool sessionClass_HasProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return false;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name)) {

			/* found property */
			return true;
		}
		session_property++;
	}

	/* TODO: need to handle read-only static values here too? (e.g. "uuid") */

	return false;
}

/**
 * sessionClass_GetProperty
 * \todo	Read-Only static properties?
 */
JSValueRef sessionClass_GetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return NULL;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name)) {

			/* found property */
			return session_property->getfunc(ctx, object);
		}
		session_property++;
	}

	/* TODO: need to handle read-only static values here too? (e.g. "uuid") */

	return NULL;
}

bool sessionClass_SetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef value, JSValueRef* exception)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return true;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name) &&
		    session_property->setfunc && !(session_property->flags & kJSPropertyAttributeReadOnly))
		{
			/* found property (and it's writable too) */
			return session_property->setfunc(ctx, object, value, exception);
		}
		session_property++;
	}

	return true;
}


/**
 * init_session_object
 * \brief	Create session object
 */
static switch_status_t init_session_object(JSContextRef context, switch_core_session_t *session)
{
	struct objectMethod *session_method = sessionMethods;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name, string;
	JSObjectRef object, method;
	JSValueRef value;

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "Creating session object for session %s\n",
				switch_core_session_get_uuid(session));

	/* session object */
	name = JSStringCreateWithUTF8CString("session");
	object = JSObjectMake(context, create_sessionClass(context), NULL);
	JSObjectSetProperty(context, globalObject, name, object, kJSPropertyAttributeNone, NULL);
	JSStringRelease(name);

	/* Setup non-static methods and properties (?) */
	JSObjectSetPrivate(object, session);

	while(session_method && session_method->name) {
		if(session_method->func) {

			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE,
					"Creating non-static method \"%s\" for session object\n",
					session_method->name);

			name = JSStringCreateWithUTF8CString(session_method->name);
			method = JSObjectMakeFunctionWithCallback(context, name, session_method->func);
			JSObjectSetProperty(context, object, name, method, kJSPropertyAttributeNone, NULL);
			JSStringRelease(name);
		}
		session_method++;
	}

	/* Read-Only value properties that never change
	 * during the lifetime of a session (e.g. its UUID)
	 *
	 * Things like "state" will change and are handled
	 * via the Has- / Get- and SetProperty helpers
	 * (and are defined in the "sessionProperties" array)
	 */

	/* uuid */
	name = JSStringCreateWithUTF8CString("uuid");
	string = JSStringCreateWithUTF8CString(switch_core_session_get_uuid(session));
	value  = JSValueMakeString(context, string);
	JSObjectSetProperty(context, object, name, value, kJSPropertyAttributeReadOnly, NULL);
	JSStringRelease(name);
	JSStringRelease(string);

	return SWITCH_STATUS_SUCCESS;
}

/******************************************************************************
 *
 *
 */

/* TODO: replace by switch_*-based function */
static char* createStringWithContentsOfFile(const char* fileName)
{
    char* buffer;

    size_t buffer_size = 0;
    size_t buffer_capacity = 1024;
    buffer = (char*)malloc(buffer_capacity);

    FILE* f = fopen(fileName, "r");
    if (!f) {
        fprintf(stderr, "Could not open file: %s\n", fileName);
        return 0;
    }

    while (!feof(f) && !ferror(f)) {
        buffer_size += fread(buffer + buffer_size, 1, buffer_capacity - buffer_size, f);
        if (buffer_size == buffer_capacity) { // guarantees space for trailing '\0'
            buffer_capacity *= 2;
            buffer = (char*)realloc(buffer, buffer_capacity);
        }
    }
    fclose(f);
    buffer[buffer_size] = '\0';

    return buffer;
}

/**
 * print_cb
 * \brief	Logging callback
 */
static JSValueRef print_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount,
		 const JSValueRef arguments[],
		 JSValueRef* exception)
{
	if (argumentCount > 0) {
		char tmp[1024];
		JSStringRef message;
		size_t size;
		int loglevel = SWITCH_LOG_NOTICE;


		if(argumentCount == 1) {
			message = JSValueToStringCopy(context, arguments[0], NULL);
		} else {
			JSStringRef level;

			level   = JSValueToStringCopy(context, arguments[0], NULL);
			message = JSValueToStringCopy(context, arguments[1], NULL);

			JSStringGetUTF8CString(level, tmp, sizeof(tmp)-1);

			if(!strcasecmp(tmp, "info")) {
				loglevel = SWITCH_LOG_INFO;
			}
			else if(!strcasecmp(tmp, "notice")) {
				loglevel = SWITCH_LOG_NOTICE;
			}
			else if(!strcasecmp(tmp, "console")) {
				loglevel = SWITCH_LOG_CONSOLE;
			}
			else if(!strcasecmp(tmp, "error")) {
				loglevel = SWITCH_LOG_ERROR;
			}
			else if(!strcasecmp(tmp, "crit")) {
				loglevel = SWITCH_LOG_CRIT;
			}
			else if(!strcasecmp(tmp, "alert")) {
				loglevel = SWITCH_LOG_ALERT;
			}
			else if(!strcasecmp(tmp, "warning")) {
				loglevel = SWITCH_LOG_WARNING;
			}
			else {
				loglevel = SWITCH_LOG_DEBUG;
			}
			JSStringRelease(level);
		}

		size = JSStringGetMaximumUTF8CStringSize(message);
		JSStringGetUTF8CString(message, tmp, ((size < sizeof(tmp)) ? size : sizeof(tmp)));

		switch_log_printf(SWITCH_CHANNEL_LOG, loglevel, "%s\n", tmp);
		JSStringRelease(message);
	}
	return JSValueMakeUndefined(context);
}



SWITCH_STANDARD_API(jsapi_function)
{
	JSGlobalContextRef context;
	JSObjectRef globalObject;
	JSValueRef  exception = NULL, result;
	JSStringRef script;
	switch_status_t retval;

	if(!cmd || switch_strlen_zero(cmd)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Not enough arguments\n");
		return SWITCH_STATUS_GENERR;
	}
	/* TODO: parse args and extract script filename */


	/*
	 * Create new Squirrelfish context
	 */
	context = JSGlobalContextCreate(NULL);
	globalObject = JSContextGetGlobalObject(context);

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "Context and global object created\n");

	/*
	 * Create global functions and classes
	 */
	retval = init_functions(context);
	if(retval != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create global functions\n");
		goto out;
	}

	retval = init_classes(context);
	if(retval != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create global classes\n");
		goto out;
	}

	/*
	 * Assign pointer to session (or NULL if there is none)
	 */
	if(session) {
		init_session_object(context, session);
	} else {
		script = JSStringCreateWithUTF8CString("var session = null;");
		result = JSEvaluateScript(context, script, NULL, NULL, 1, &exception);

		if(JSValueIsUndefined(context, result))
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "session setting script executed successfully\n");

		JSStringRelease(script);
	}

	/*
	 * Load script
	 */
	char *scriptUTF8 = createStringWithContentsOfFile("/opt/freeswitch/scripts/test.js");
	if(!scriptUTF8) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to load script\n");
		goto out;
	}

	script = JSStringCreateWithUTF8CString(scriptUTF8);
        result = JSEvaluateScript(context, script, NULL, NULL, 1, &exception);
	if(JSValueIsUndefined(context, result)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Script executed successfully\n");
	} else {
		char tmp[1024];
		size_t size;

		JSStringRef exceptionString = JSValueToStringCopy(context, exception, NULL);

		size = JSStringGetMaximumUTF8CStringSize(exceptionString);
		JSStringGetUTF8CString(exceptionString, tmp, ((size < sizeof(tmp)) ? size : sizeof(tmp)));

		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Script failed: %s\n", tmp);

		JSStringRelease(exceptionString);
	}

	JSStringRelease(script);
	free(scriptUTF8);

out:
	globalObject = NULL;
	JSGlobalContextRelease(context);
	JSGarbageCollect(context);

	return SWITCH_STATUS_SUCCESS;
}


SWITCH_MODULE_LOAD_FUNCTION(mod_squirrelfish_load)
{
	switch_api_interface_t *api_interface;

	/* connect my internal structure to the blank pointer passed to me */
	*module_interface = switch_loadable_module_create_module_interface(pool, modname);

	SWITCH_ADD_API(api_interface, "jsrun", "Run a JavaScript script", jsapi_function, "jsapi <script> [additional_vars [...]]");

	/* indicate that the module should continue to be loaded */
	return SWITCH_STATUS_SUCCESS;
}


/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:nil
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 expandtab:
 */
